/*=============================================================================
	Collision Cacus header
=============================================================================*/

#ifndef _INC_XC_COL
#define _INC_XC_COL

#define MAX_OCTREE_DEPTH 6
#define OCTREE_GRID_ARRAY 7
#define TWO_EXP_OGA 64 //2^OCTREE_GRID_ARRAY

#define MIN_ACTORS_FOR_SUBNODE 3


class FPrecomputedRay;
struct FCacusCylActorInfo;

class XC_ENGINE_API AXC_PrimitiveActor : public AActor
{
public:
    BITFIELD bNoLineCheck:1 GCC_PACK(4);
    BITFIELD bNoPointCheck:1;
    BITFIELD bNoEncroachmentCheck:1;
    BITFIELD bNoExtentCheck:1;
    BITFIELD bForceZeroExtent:1;
    BITFIELD bFallBackToCylinder:1;
    BITFIELD bBackTrace:1;
    INT ActorChannel GCC_PACK(4);

	virtual UBOOL RayTrace( FMemStack& Mem, const FPrecomputedRay& Ray, FVector* HitLoc, FVector* HitNorm)
	{	return 0;	}
	virtual UBOOL InContact( FMemStack& Mem, FVector InLoc, FVector Extent, FCheckResult* Hit)
	{	return 0;	}
	
    DECLARE_CLASS(AXC_PrimitiveActor,AActor,0,XC_Engine)
    NO_DEFAULT_CONSTRUCTOR(AXC_PrimitiveActor)
};

class XC_ENGINE_API AXC_PrimitiveSphere : public AXC_PrimitiveActor
{
public:
    BITFIELD bHollow:1 GCC_PACK(4);
    BITFIELD bNoHollowFrontHits:1;
	
	//AXC_PrimitiveActor interface
	UBOOL RayTrace( FMemStack& Mem, const FPrecomputedRay& Ray, FVector* HitLoc, FVector* HitNorm);
	UBOOL InContact( FMemStack& Mem, FVector InLoc, FVector Extent, FCheckResult* Hit);
	
    DECLARE_CLASS(AXC_PrimitiveSphere,AXC_PrimitiveActor,0,XC_Engine)
    NO_DEFAULT_CONSTRUCTOR(AXC_PrimitiveSphere)
};

class XC_ENGINE_API AXC_PrimitivePlane : public AXC_PrimitiveActor
{
public:
    BITFIELD bBlockOutgoing:1 GCC_PACK(4);
    BITFIELD bBlockIncoming:1;
    BITFIELD bUseMyLocation:1;
    BITFIELD bUseMyRotation:1;
    BITFIELD bSolidifyBehindPlane:1;
    FPlane iPlane GCC_PACK(4);
	FLOAT iHAxis;
    FVector CachedLocation;
    FRotator CachedRotation;
	
	//AXC_PrimitiveActor interface
	UBOOL RayTrace( FMemStack& Mem, const FPrecomputedRay& Ray, FVector* HitLoc, FVector* HitNorm);
	UBOOL InContact( FMemStack& Mem, FVector InLoc, FVector Extent, FCheckResult* Hit);
	
	//AXC_PrimitivePlane interface
	void VerifyInternals();
	
    DECLARE_CLASS(AXC_PrimitivePlane,AXC_PrimitiveActor,0,XC_Engine)
    NO_DEFAULT_CONSTRUCTOR(AXC_PrimitivePlane)
};

struct FFacePoints
{
	INT Vertex[3];
	FLOAT HNorm; //Norm of X,Y components of plane normal
	
	inline FFacePoints& operator=(const FMeshTri& Tri)
	{
		Vertex[0] = Tri.iVertex[0];
		Vertex[1] = Tri.iVertex[1];
		Vertex[2] = Tri.iVertex[2];
		return *this;
	}
	
	inline FFacePoints& operator=(const FMeshFace& Face)
	{
		Vertex[0] = Face.iWedge[0];
		Vertex[1] = Face.iWedge[1];
		Vertex[2] = Face.iWedge[2];
		return *this;
	}
	
};

class XC_ENGINE_API AXC_PrimitiveMesh : public AXC_PrimitiveActor
{
public:
    TArray<FLOAT> SNorm;
    FVector4* TVerts;
    INT TVertsCount;
    FVector4* TPlanes;
    INT TPlanesCount;
	FFacePoints* TVertices;
	INT* TPlaneDots;

	//UObject interface
	void Destroy(); //Deallocate cache
	
	//AXC_PrimitiveActor interface
	UBOOL RayTrace( FMemStack& Mem, const FPrecomputedRay& Ray, FVector* HitLoc, FVector* HitNorm);
	UBOOL InContact( FMemStack& Mem, FVector InLoc, FVector Extent, FCheckResult* Hit);

	//AXC_PrimitiveMesh interface
	UBOOL VerifyInternals();
	void SetupMesh();
	FCheckResult* ZeroExtentCheck( FMemStack& Mem, FVector4* LocalStart, FVector4* LocalEnd, FVector4* Dir);
	FCheckResult* ExtentCheck( FMemStack& Mem, FVector4* LocalStart, FVector4* LocalEnd, FVector4* Dir, const FVector4* Extent);

	
    DECLARE_CLASS(AXC_PrimitiveMesh,AXC_PrimitiveActor,0,XC_Engine)
    NO_DEFAULT_CONSTRUCTOR(AXC_PrimitiveMesh)
};


MS_ALIGN(16) class FPrecomputedRay
{
public:
	FVector4 Org;
	FVector4 End;
	FVector4 Dir;
	FVector4 Inv;
	FVector4 Extent;
	FBox ExtendedBox; //For grid interaction and quick rejects
	FLOAT Dist;
	UBOOL(FPrecomputedRay::*TraceFunc)( FCacusCylActorInfo*, FVector4*, FVector4*, AXC_PrimitiveActor* &);

	FPrecomputedRay()	{};
	FPrecomputedRay( const FVector& A, const FVector& B, const FVector& TExtent);


	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X <= Container.Max.X) && (Sample.Y <= Container.Max.Y) && (Sample.Z <= Container.Max.Z);
	}

	INT IntersectsBox( const FBox& Box);

	FORCEINLINE UBOOL HitsGCylActor( FCacusCylActorInfo* AInfo, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive);
	UBOOL HitsVCylActor( FCacusCylActorInfo* AInfo, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive);
	UBOOL HitsHCylActor( FCacusCylActorInfo* AInfo, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive);
	UBOOL HitsCylActor(  FCacusCylActorInfo* AInfo, FVector4* HitLocation, FVector4* HitNormal, AXC_PrimitiveActor* & Primitive);
} GCC_ALIGN(16);

struct FCacusActorInfo
{
	AActor* Actor;
	INT ObjIndex; //Avoid querying the reference for stability purposes
	INT TopDepth;
	FVector Location;

	FCacusActorInfo( AActor* NewActor)
	:	Actor(NewActor)
	,	ObjIndex(NewActor->GetIndex())
	,	TopDepth(0) //Increase when possible to add into new node
	,	Location( NewActor->Location)
	{}

	inline UBOOL IsValid() const
	{
		return (UObject::GetIndexedObject(ObjIndex) == Actor) && Actor->bCollideActors && !Actor->bDeleteMe;
	}
};

struct FCacusPrimActorInfo : public FCacusActorInfo //Generic primitive version
{
	FBox Box;

	FCacusPrimActorInfo( AActor* NewActor, FBox NewBox)
	:	FCacusActorInfo(NewActor)
	,	Box(NewBox)
	{
		Location = (NewBox.Min+NewBox.Max)*0.5;
	}
};

struct FCacusCylActorInfo : public FCacusActorInfo //Optimial cylinder version
{
	FLOAT Radius;
	FLOAT Height;
	INT Padding[5]; //Allows branchless implemenation of memory operations
	
	FCacusCylActorInfo( AActor* NewActor)
	:	FCacusActorInfo( NewActor)
	,	Radius( NewActor->CollisionRadius)
	,	Height( NewActor->CollisionHeight)
	{}
	
	//Could be easily ported to SSE
	inline UBOOL TouchesBox( const FBox& Box)
	{
		return (Location.X+Radius >= Box.Min.X && Location.X-Radius <= Box.Max.X 
			&&  Location.Y+Radius >= Box.Min.Y && Location.Y-Radius <= Box.Max.Y 
			&&  Location.Z+Height >= Box.Min.Z && Location.Z-Height <= Box.Max.Z );
	}
};


class XC_ENGINE_API FCollisionOctree
{
public:
	FCollisionCacus* Master;
	FCollisionOctree* Parent;
	FCollisionOctree* ReducedChildren[8]; //For fast operations
	FBox BoundingBox;
	BYTE CurDepth;
	BYTE ChildIndex;
	BYTE ChildCount;
	BYTE CleanupTimer;
	BYTE XPos, YPos, ZPos;
	INT MIndex; //Index on master array
	INT TotalChildren;
	TArray<FCacusCylActorInfo> Actors;
	TArray<FCacusPrimActorInfo> BrushMovers; //Brush && bIsMover
	//FCollisionCacus uses this
	FCollisionOctree()
	:	CurDepth(0)
	,	ChildIndex(0)
	,	ChildCount(0)
	,	CleanupTimer(20)
	,	Actors(0)
	,	BrushMovers(0)
	,	XPos(0)	,	YPos(0)	,	ZPos(0)
	,	TotalChildren(0)
	,	BoundingBox( FVector(-32768,-32768,-32768), FVector(32768,32768,32768))
	{
		appMemzero( &Parent, 4*9);
	}
	FCollisionOctree( FCollisionOctree* InParent, INT* XYZ);
	~FCollisionOctree();
	virtual void Exit();

	void Cleanup();
	FCollisionOctree* SlowRemoveActor( AActor* Other);
	FCacusActorInfo* FindActor( FName ByName);
	FIteratorActorList* RadiusQuery( FMemStack& Mem, const FVector& Location, const FLOAT& Radius, const DWORD& ExtraNodeFlags, FIteratorActorList* PrevLink );
	FCheckResult* FastPointQuery( FMemStack& Mem, const FVector& Location, const FVector& Extent, const DWORD& ExtraNodeFlags, FCheckResult* PrevLink);
	FCheckResult* MEncroachQuery( FMemStack& Mem, AActor* Actor, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* FastAEncroachQuery( FMemStack& Mem, AActor* Actor, const FVector& Location, const FBox& ActorBox, FCheckResult* PrevLink=NULL);
	FCheckResult* FastLineQuery
	(
	 FMemStack& Mem
	, const FVector& End
	, const FVector& Start
	, const FVector& Extent
	, const BYTE& ExtraNodeFlags
	, FPrecomputedRay& Ray
	, FCheckResult* PrevLink
	);
	inline UBOOL FitsBox( const FBox& Sample, const FBox& Container)
	{
		return (Sample.Min.X >= Container.Min.X) && (Sample.Min.Y >= Container.Min.Y) && (Sample.Min.Z >= Container.Min.Z)
			&& (Sample.Max.X <= Container.Max.X) && (Sample.Max.Y <= Container.Max.Y) && (Sample.Max.Z <= Container.Max.Z);
	}
	inline UBOOL FitsBox( const FVector& Sample, const FBox& Container)
	{
		return (Sample.X >= Container.Min.X) && (Sample.Y >= Container.Min.Y) && (Sample.Z >= Container.Min.Z)
			&& (Sample.X < Container.Max.X) && (Sample.Y < Container.Max.Y) && (Sample.Z < Container.Max.Z);
	}
	inline UBOOL FitsBox( AActor* Actor, const FBox& Container)
	{
		return (Actor->Location.X - Actor->CollisionRadius >= Container.Min.X) && (Actor->Location.Y - Actor->CollisionRadius >= Container.Min.Y) && (Actor->Location.Z - Actor->CollisionHeight >= Container.Min.Z)
			&& (Actor->Location.X + Actor->CollisionRadius < Container.Max.X) && (Actor->Location.Y + Actor->CollisionRadius < Container.Max.Y) && (Actor->Location.Z + Actor->CollisionHeight < Container.Max.Z);
	}
};

//====================================
// Three dimension grid
//==========

//V451 is cancer, i need to work around it by statically allocating a struct
struct FGridData
{
	INT Position[OCTREE_GRID_ARRAY];

	FCollisionOctree* Grid0;
#if OCTREE_GRID_ARRAY > 0
	FCollisionOctree* Grid1[2][2][2];
#endif
#if OCTREE_GRID_ARRAY > 1
	FCollisionOctree* Grid2[4][4][4];
#endif
#if OCTREE_GRID_ARRAY > 2
	FCollisionOctree* Grid3[8][8][8];
#endif
#if OCTREE_GRID_ARRAY > 3
	FCollisionOctree* Grid4[16][16][16];
#endif
#if OCTREE_GRID_ARRAY > 4
	FCollisionOctree* Grid5[32][32][32];
#endif
#if OCTREE_GRID_ARRAY > 5
	FCollisionOctree* Grid6[64][64][64];
#endif
#if OCTREE_GRID_ARRAY > 6
	FCollisionOctree* Grid7[128][128][128];
#endif
#if OCTREE_GRID_ARRAY > 7
	#error NO_LONGER_SUPPORTED
	Error;
#endif
	INT MaxSlots; //Absolute access

	FCollisionOctree* FullList( INT RawIdx)
	{
		if ( RawIdx < 0 || RawIdx >= MaxSlots )
			return NULL;
		return (&Grid0)[RawIdx];
	}

	FGridData()
	{
		MaxSlots = ((DWORD)&MaxSlots - (DWORD)&Grid0) >> 2; //Divide by 4
		Position[0] = 0;
		for ( INT i=1 ; i<OCTREE_GRID_ARRAY ; i++ )
		{
			INT PrevSize = (1 << (i-1));
			PrevSize = PrevSize * PrevSize * PrevSize;
			Position[i] = Position[i-1] + PrevSize;
			INT NewSize = (1 << i);
			FCollisionOctree** TmpGrid = &Grid0;
			TmpGrid += Position[i];
			appMemzero( TmpGrid, NewSize * NewSize * NewSize * 4);
		}
		check( Position[1] - Position[0] == 1 );
		check( Position[2] - Position[1] == 8 );
		check( Position[3] - Position[2] == 64 );
		check( Position[4] - Position[3] == 512 );
		check( Position[5] - Position[4] == 4096 );
		check( Position[6] - Position[5] == 32768 );
	}
};


class XC_ENGINE_API FNodeGrid3
{
public:
	INT GridSize;
	FCollisionOctree** Grid;
	FVector Substract; //Substract this when calculating grid position, should be identical in all grids?
	FVector BoxSize; //Dividing a vector by this (post substraction) will give coordinates!
	FVector Divider; //Precalculated version

	FNodeGrid3()
	:	GridSize(0)
	,	Grid(NULL)
	{};

	void SetBaseGrid( FCollisionOctree* Base);
	void SetGrid( FNodeGrid3* Upper);
	void Register( FCollisionOctree* In);
	void Remove( FCollisionOctree* In);
	INT GetIdx( const FVector& Pos);
	void GetXYZ( const FVector& Pos, INT* XYZ);
	inline FCollisionOctree* Get( INT RawIdx)
	{
		return Grid[RawIdx];
	}
	inline FCollisionOctree* GetByXYZ( INT X, INT Y, INT Z)
	{
		return Grid[X + Y*GridSize + Z*GridSize*GridSize];
	}
	inline FBox GetBox( INT X, INT Y, INT Z)
	{
		FVector Min = Substract + FVector( BoxSize.X * X, BoxSize.Y * Y, BoxSize.Z * Z);
		return FBox( Min, Min+BoxSize);
	}
};

class XC_ENGINE_API FCollisionCacus : public FCollisionHashBase , public FCollisionOctree
{
public:
	FGridData* GridData;
	FCoords PrecomputedCoords;
	FNodeGrid3 DepthGrid[OCTREE_GRID_ARRAY]; //Zero isn't used, we're doing this for quick CurDepth access

	FCollisionCacus( ULevel* Level)
	:	FCollisionOctree()
	,	GridData(NULL)
	{
		MIndex = 0;
		Master = this;
		PrecomputedCoords.ZAxis = FVector(0.f,0.f,1.f);
		appMemzero( &DepthGrid, sizeof(class FNodeGrid3) * OCTREE_GRID_ARRAY );
		//Adjust to substractive level
		if ( Level && Level->Model && Level->Model->Points.Num() )
		{
			BoundingBox = FBox( &Level->Model->Points(0), Level->Model->Points.Num() );
			BoundingBox.Min -= FVector( 128, 128, 128);
			BoundingBox.Max += FVector( 128, 128, 128);
		}
		SetupGrids();
	};
	~FCollisionCacus();
	void Exit();

	void SetupGrids();
	void Tick();
	void AddActor( AActor *Actor );
	void RemoveActor( AActor *Actor );
	FCheckResult* ActorLineCheck( FMemStack& Mem, FVector End, FVector Start, FVector Extent, BYTE ExtraNodeFlags );
	FCheckResult* ActorPointCheck( FMemStack& Mem, FVector Location, FVector Extent, DWORD ExtraNodeFlags );
	FCheckResult* ActorRadiusCheck( FMemStack& Mem, FVector Location, FLOAT Radius, DWORD ExtraNodeFlags );
	FCheckResult* ActorEncroachmentCheck( FMemStack& Mem, AActor* Actor, FVector Location, FRotator Rotation, DWORD ExtraNodeFlags );
	void CheckActorNotReferenced( AActor* Actor );
};
#endif

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
